//
// Simple demo of z-fighting and using of glPolygonOffset to cope with it
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include	"libExt.h"

#include	<glut.h>
#include	<stdio.h>
#include	<stdlib.h>

#include	"Vector3D.h"
#include	"Plane.h"

Vector3D	eye   ( -3, 0, -2.5 );		// camera position

Vector3D	p [4];						// rectangle
Vector3D	q [3];

bool	useOffset = true;

void init ()
{
	glClearColor ( 0.0, 0.0, 0.0, 1.0 );
	glEnable     ( GL_DEPTH_TEST );
	glDepthFunc  ( GL_LEQUAL );

	glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
	glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );

	glPolygonOffset ( -1, -1 );
}

void	drawPoly ( Vector3D * p, int n )
{
	glBegin ( GL_POLYGON );

	for ( int i = 0; i < n; i++ )
		glVertex3fv ( p [i] );

	glEnd ();
}

void display ()
{
	glClear       ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	glColor3f ( 1, 1, 0 );

	drawPoly ( p, 4 );

	if ( useOffset )
		glEnable ( GL_POLYGON_OFFSET_FILL );

	glColor3f ( 0, 0, 1 );

	drawPoly ( q, 3 );

	glDisable ( GL_POLYGON_OFFSET_FILL );

	glutSwapBuffers ();
}

void reshape ( int w, int h )
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
   gluLookAt      ( eye.x, eye.y, eye.z,	// eye
					0, 0, 0,				// center
					0.0, 0.0, 1.0 );		// up
}

void key ( unsigned char key, int x, int y )
{
	if ( key == 27 || key == 'q' || key == 'Q' )
    	exit ( 0 );

	if ( key == '1' )
		useOffset = false;
	else
	if ( key == '2' )
		useOffset = true;

	glutPostRedisplay ();
}

inline	float f ( float x, float y )
{
	return x * 0.03f - y * 0.1f;
}

int main ( int argc, char** argv )
{
	printf ( "Starting.\n" );

								// initialise glut
	glutInit            ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
	glutInitWindowSize  ( 500, 500 );


								// create window
	int	glWin = glutCreateWindow ( "OpenGL z-fighting demo. Press \'1\' and \'2\' to turn polygon offset on/off" );

								// register handlers
	glutDisplayFunc  ( display );
	glutReshapeFunc  ( reshape );
	glutKeyboardFunc ( key );

	init ();

	printfInfo ();

	initExtensions ();

								// create quad (not parallel to coordinate planes)
	p [0] = Vector3D ( -1, -1, f ( -1, -1 ) );
	p [1] = Vector3D (  1, -1, f ( -1,  1 ) );
	p [2] = Vector3D (  1,  1, f (  1,  1 ) );
	p [3] = Vector3D ( -1,  1, f ( -1,  1 ) );

								// compute it's plane
	Plane	plane ( p [0], p [1], p [2] );

								// now create triangle right on it
	q [0] = plane.project ( Vector3D ( -0.8,  -0.79, 0 ) );
	q [1] = plane.project ( Vector3D ( 0.21,  -0.84, 0 ) );
	q [2] = plane.project ( Vector3D ( 0.92,   0.89, 0 ) );

	glutMainLoop ();

	return 0;
}
